from datetime import datetime
from enum import Enum
from uuid import uuid4

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models


def get_sorted_enum_types(enum_class):
    """
    Returns a sorted tuple list of the enum's values sutable for a CharField's
    ``choices`` argument
    :param enum_class: A class that inherits Enum
    :return: A tuple of tuples ((element.value, element.name), ...)
    """
    choices = [(element.value, element.name) for element in list(enum_class)]
    choices.sort()
    return tuple(choices)


class ActivityGroupStatus(Enum):
    """
    Enum for the possible status states of an activity group
    """
    ACTIVE = 'ACTIVE'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'


class ActivityStatus(Enum):
    """
    Enum for the possible status states of an activity
    """
    PENDING = 'PENDING'
    ACTIVE = 'ACTIVE'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'


class ActivityGroup(models.Model):
    """
    General-purpose model to track state of a process
    """
    uuid = models.UUIDField(db_index=True, default=uuid4, editable=False, null=True)
    name = models.CharField(max_length=255, blank=False, null=False)
    time_created = models.DateTimeField(auto_now_add=True, db_index=True)
    time_finished = models.DateTimeField(null=True, db_index=True)
    # an optional context object for filtering activity groups by related sets
    context_object_type = models.ForeignKey(ContentType, null=True, on_delete=models.PROTECT)
    context_object_id = models.PositiveIntegerField(null=True)
    context_object = GenericForeignKey('context_object_type', 'context_object_id')
    status = models.CharField(
        max_length=255,
        choices=get_sorted_enum_types(ActivityGroupStatus),
        default=ActivityGroupStatus.ACTIVE.value,
        db_index=True
    )
    error_message = models.TextField(blank=True, null=True)
    # attributes to track progress of this process
    # this is especially helpful when this process can't be broken-down into individual activities
    steps_total = models.IntegerField(null=True)
    steps_complete = models.IntegerField(null=True)

    def __str__(self):
        return self.name

    def finish(self, status=ActivityGroupStatus.SUCCESS.value):
        ActivityGroup.objects.filter(id=self.id).update(status=status, time_finished=datetime.utcnow())
        return self

    def success(self):
        return self.finish(status=ActivityGroupStatus.SUCCESS.value)

    def failure(self, message):
        ActivityGroup.objects.filter(id=self.id).update(
            status=ActivityGroupStatus.FAILURE.value, time_finished=datetime.utcnow(), error_message=message)
        return self


class Activity(models.Model):
    """
    Model to track the state of a single activity belonging to a process
    """
    name = models.CharField(max_length=255, blank=False, null=False)
    description = models.TextField(blank=False, null=False)
    time_created = models.DateTimeField(auto_now_add=True, db_index=True)
    time_updated = models.DateTimeField(auto_now=True, db_index=True)
    time_finished = models.DateTimeField(null=True, db_index=True)
    group = models.ForeignKey(ActivityGroup, null=True, related_name='activities', on_delete=models.CASCADE)
    status = models.CharField(
        max_length=255,
        default=ActivityStatus.PENDING.value,
        choices=get_sorted_enum_types(ActivityStatus),
        db_index=True
    )
    error_message = models.TextField(blank=True, null=True)

    def finish(self, status=ActivityStatus.SUCCESS.value, error_message=None):
        self.status = status
        self.time_finished = datetime.utcnow()
        self.error_message = error_message
        self.save(update_fields=[
            'status',
            'time_finished',
            'error_message'
        ])
        return self

    def active(self):
        self.status = ActivityStatus.ACTIVE.value
        self.error_message = None
        self.time_finished = None
        self.save(update_fields=[
            'status',
            'time_finished',
            'error_message'
        ])
        return self

    def success(self):
        return self.finish(status=ActivityStatus.SUCCESS.value)

    def failure(self, error_message):
        return self.finish(status=ActivityStatus.FAILURE.value, error_message=error_message)



# import multiprocessing to avoid this bug (http://bugs.python.org/issue15881#msg170215)
import multiprocessing
assert multiprocessing
import re
from setuptools import setup, find_packages


def get_version():
    """
    Extracts the version number from the version.py file.
    """
    VERSION_FILE = 'ambition_utils/version.py'
    mo = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', open(VERSION_FILE, 'rt').read(), re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError('Unable to find version string in {0}.'.format(VERSION_FILE))


def get_lines(file_path):
    return open(file_path, 'r').read().split('\n')


install_requires = get_lines('requirements/requirements.txt')
tests_require = get_lines('requirements/requirements-testing.txt')


setup(
    name='ambition-utils',
    version=get_version(),
    description='Various utility packages used across Ambition projects.',
    long_description=open('README.rst').read(),
    url='https://github.com/ambitioninc/ambition-utils',
    author='Wes Okes',
    author_email='wes.okes@gmail.com',
    keywords='django, database, query, sql, postgres, upsert',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
    ],
    license='MIT',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={'dev': tests_require},
    test_suite='run_tests.run_tests',
    include_package_data=True,
)



from django.db import models
from timezone_field.fields import TimeZoneField as BaseTimeZoneField

from ambition_utils.fields import TimeZoneField


class FakeModel(models.Model):
    name = models.CharField(max_length=50)
    cast_time_zone_field = TimeZoneField(default='utc', null=True)
    no_cast_time_zone_field = BaseTimeZoneField(default='utc', null=True)



from celery import Celery
from django.test import TestCase
from unittest.mock import patch
from uuid import uuid4

from ambition_utils.activity.tasks import ActivityManagedTaskMixin, track_activity
from ambition_utils.activity.models import Activity, ActivityGroup, ActivityStatus, ActivityGroupStatus

test_application = Celery('task_tests')


class TestTask(ActivityManagedTaskMixin, test_application.Task):

    def run(self):
        self.first()
        self.second()
        self.third()
        self.fourth()
        self.fifth()

    @track_activity
    def first(self):
        pass

    @track_activity
    def second(self):
        pass

    @track_activity
    def third(self):
        pass

    @track_activity
    def fourth(self):
        pass

    def fifth(self):
        pass


class ActivityManagedTaskMixinTest(TestCase):

    def test_activity_tracking(self):
        TestTask(uuid=uuid4()).run()
        self.assertEqual(4, Activity.objects.count())
        successful_activities = Activity.objects.filter(status=ActivityStatus.SUCCESS.value)
        self.assertEqual(4, successful_activities.count())

        # ensure that there is only one activity gruop
        activity_group = ActivityGroup.objects.get()
        # ensure that the group is marked as successful
        self.assertEqual(ActivityGroupStatus.SUCCESS.value, activity_group.status)
        self.assertEqual('TestTask', activity_group.name)

    @patch.object(TestTask, 'second')
    def test_failed_activity(self, mock_second_activity):
        # make the second activity fail
        mock_second_activity.side_effect = Exception('wat')
        mock_second_activity.__name__ = 'foo'
        task = TestTask(uuid=uuid4())
        with self.assertRaises(Exception):
            task.run()
        self.assertEqual(1, Activity.objects.filter(status=ActivityStatus.SUCCESS.value).count())
        self.assertEqual(1, Activity.objects.filter(status=ActivityStatus.FAILURE.value).count())
        self.assertEqual(2, Activity.objects.filter(status=ActivityStatus.PENDING.value).count())

        # ensure that there is only one activity gruop
        activity_group = ActivityGroup.objects.get()
        # ensure that the group is marked as successful
        self.assertEqual(ActivityGroupStatus.FAILURE.value, activity_group.status)
        self.assertEqual('wat', activity_group.error_message)

    def test_activity_group_name(self):
        self.assertEqual(
            ActivityManagedTaskMixin().get_activity_group_name(),
            'ActivityManagedTaskMixin'
        )

        class TestTask(ActivityManagedTaskMixin):
            def get_activity_group_name(self):
                return 'test name'
        mixin = TestTask()
        self.assertEqual(mixin.get_activity_group_name(), 'test name')



# Generated by Django 2.0.6 on 2018-06-04 19:00

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PostgresLock',
            fields=[
                ('name', models.CharField(max_length=255, primary_key=True, serialize=False, unique=True)),
                ('time', models.DateTimeField(default=None)),
            ],
        ),
    ]



# Generated by Django 2.2.4 on 2019-10-30 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postgres_lock', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='postgreslock',
            name='name',
            field=models.CharField(max_length=512, primary_key=True, serialize=False, unique=True),
        ),
    ]



import wrapt

from copy import deepcopy, copy
from django import forms

from django.core.exceptions import ValidationError
from django.db import transaction


class NestedFormConfig(object):
    """
    Defines how a nested form is handled in the context of another form. Any form class using a subclass of
    NestedFormMixinBase can define a nested_form_configs list with instances of NestedFormConfig
    """

    def __init__(
        self,
        cls,
        key,
        required=False,
        field_prefix=None,
        required_key=None,
        pre=False,
        post=False,
        error_messages=None
    ):
        """
        Sets all default values

        :param cls: Any form class reference

        :param key: The dict key that will be used when passing around form arguments during processing.
                    The value of this dict will be the return value of the save method
        :type key: str

        :param required: Determines if this form's fields are always required
        :type required: bool

        :param field_prefix: Optional field for when multiple instance of the same form are nested.
                             This allows the submitted data to have a prefix before each form field name.
        :type field_prefix: str

        :param required_key: Optional form field that causes the form to be required based on its presence.
        :type required_key: str or None

        :param pre: Flag to indicate that the nested form should be processed before the parent form save method.
                    This allows the nested form's save value to be available in the save method
                    arguments keyed off of the key param.
        :type pre: bool

        :param post: Flag to indicate that the nested form should be processed after the parent form save method.
                    This allows the parent form's save value to be available in the post save method arguments.
        :type post: bool

        :param error_messages: Any error messages to override. This should be a multi dict with the top
                               level being the field name and its value being a dictionary of error_type => message
        :type error_messages: dict
        """

        # Store all the properties onto the class
        self.cls = cls
        self.key = key
        self.required = required
        self.field_prefix = field_prefix
        self.required_key = required_key
        self.pre = pre
        self.post = post
        self.error_messages = error_messages or {}
        self.instance = None

        # Assert that we have a class and key objects
        assert self.cls
        assert self.key

    def set_instance(self, *args, **kwargs):
        """
        Set the instance of the created form
        """

        # Create the instance with the passed arguments
        self.instance = self.cls(*args, **kwargs)

        # Check if there are any custom error message overrides defined
        if self.error_messages:

            # Copy the error messages from the form class to the form instance so it doesn't alter the class
            for field_name, field in self.instance.fields.items():
                self.instance.fields[field_name].error_messages = deepcopy(
                    self.instance.fields[field_name].error_messages
                )

        # Apply any custom error messages
        for field_name, messages in self.error_messages.items():
            if field_name in self.instance.fields:
                self.instance.fields[field_name].error_messages.update(messages)


class NestedFormMixin(object):
    """
    Allows a form to contain multiple and optional nested forms. The form configurations are defined in
    nested_form_configs and is defined by instances of NestedFormConfig
    """

    # A list of nested form config items that should be nested within the main form
    nested_form_configs = []

    def __init__(self, *args, **kwargs):
        # Call the parent
        super(NestedFormMixin, self).__init__(*args, **kwargs)

        # Make a copy of the original data because we will be manipulating it and
        # potentially adding additional keys from our nested forms
        self.data = copy(self.data)

        # Assert that we have a save method
        if not hasattr(self, 'save'):
            raise Exception('Base form must have a save method - {0}'.format(self.__class__))

        # Wrap the save method around the nested form save
        self._save = self.save
        self.save = self._nested_save(self._save)

        # Build a list of all nest form configs
        self.nested_forms = []

        # Keep track of form prefixes to guarantee multiple of the same form are properly prefixed
        form_prefixes = {}
        for nested_form_config in self.nested_form_configs:
            # Check if this form class already exists
            if nested_form_config.cls in form_prefixes:
                # Make sure both have a prefix value
                if not form_prefixes[nested_form_config.cls] or not nested_form_config.field_prefix:
                    raise ValidationError(
                        'Form {0} must have a field prefix'.format(nested_form_config.cls.__name__)
                    )

            # Create a link field on the main form
            # Check if a field already exists
            if nested_form_config.key in self.fields:
                raise Exception('A nested field with key {0} already exists on the base form'.format(
                    nested_form_config.key
                ))

            # Add the field to the base form so we can add errors and data for cleaning
            self.fields[nested_form_config.key] = forms.Field(required=False)

            # Build the nested form args, kwargs to init the nested form with
            nested_form_args = []
            nested_form_kwargs = {
                'data': {},
                'files': {},
            }

            # Get the prefix
            prefix = nested_form_config.field_prefix

            # Set the prefix value to the form config prefix
            form_prefixes[nested_form_config.cls] = nested_form_config.field_prefix

            # Process the form field and file keys when there is a prefix defined on the nested form
            for label, data in (('data', self.data), ('files', self.files)):
                for key, value in copy(data).items():
                    # Check if the prefix is there to replace
                    to_replace = '{0}_'.format(prefix)
                    if key.startswith(to_replace):
                        # Update the key
                        key = key.replace(to_replace, '')

                    # Apply the value to the data dict
                    nested_form_kwargs[label][key] = value

            # Get the nested form config init arguments
            nested_form_args, nested_form_kwargs = self.get_nested_form_init_args(
                nested_form_config=nested_form_config,
                args=nested_form_args,
                kwargs=nested_form_kwargs,
                base_form_args=args,
                base_form_kwargs=kwargs
            )

            # Create the form instance and pass the form data
            nested_form_config.set_instance(*nested_form_args, **nested_form_kwargs)

            # Add the form config to the list of nested form configs
            self.nested_forms.append(nested_form_config)

    @wrapt.decorator
    @transaction.atomic
    def _nested_save(self, wrapped, instance, args, kwargs):
        """
        This is a wrapper around the base forms save method.
        Handles getting additional arguments, calling save on pre-save forms, calling save_form on the parent form,
        and calling save on post-save forms. The form kwargs are passed through the whole chain and the parent
        form's save_form return value is returned from this method.

        :param wrapped: The original method
        :param instance: The instance the method is attached to
        :param args:
        :param kwargs:
        :return:
        """

        # Get all required nested forms
        required_forms = self.get_required_forms()

        # Keep track of the form responses
        responses = {}

        # Save all pre-save forms and apply the results to the base form kwargs and nested form kwargs
        for nested_form_config in required_forms:
            if nested_form_config.pre:
                responses = self._save_nested_form(
                    nested_form_config=nested_form_config,
                    base_form_args=args,
                    base_form_kwargs=kwargs,
                    responses=responses
                )

        # Save the parent form and store the result under the save key
        responses['base'] = wrapped(*args, **kwargs, **responses)

        # Save all post-save forms
        for nested_form_config in required_forms:
            if nested_form_config.post:
                responses = self._save_nested_form(
                    nested_form_config=nested_form_config,
                    base_form_args=args,
                    base_form_kwargs=kwargs,
                    responses=responses
                )

        # Call the save nested method
        self.save_nested(responses['base'], **responses)

        # Return the value from the parent form's save method
        return responses['base']

    def _save_nested_form(self, nested_form_config, base_form_args, base_form_kwargs, responses):
        # Get the args and kwargs for the nested form
        nested_form_args, nested_form_kwargs = self.get_nested_form_save_args(
            nested_form_config=nested_form_config,
            args=[],
            kwargs={},
            base_form_args=base_form_args,
            base_form_kwargs=base_form_kwargs,
            form_responses=responses
        )

        # Call the save method of the nested form
        response = nested_form_config.instance.save(*nested_form_args, **nested_form_kwargs)

        # Store the response
        responses[nested_form_config.key] = response

        # Return the responses
        return responses

    def save_nested(self, response, **kwargs):
        """
        Called after all nested forms and the main form have been saved
        :param response: The response from the main form
        :param kwargs: The nested kwargs
        """
        pass

    def get_nested_form_init_args(self, nested_form_config, args, kwargs, base_form_args, base_form_kwargs):
        """
        Given a nested form config get the init arguments, should return args, kwargs

        :param nested_form_config: The nested form config
        :param args: The args that will be passed into the init method
        :param kwargs: The kwargs that will be passed into the init method
        :param base_form_args: The args that were given to the base form
        :param base_form_kwargs: The kwargs that were given to the base form
        """
        return args, kwargs

    def get_nested_form_save_args(
        self,
        nested_form_config,
        args,
        kwargs,
        base_form_args,
        base_form_kwargs,
        form_responses
    ):
        """
        Given a nested form config get the save arguments, should return args, kwargs

        :param nested_form_config: The nested form config
        :param args: The args that will be passed into the init method
        :param kwargs: The kwargs that will be passed into the init method
        :param base_form_args: The args that were given to the base form
        :param base_form_kwargs: The kwargs that were given to the base form
        :param form_responses: Any current responses we have from saved forms
        """
        return args, kwargs

    def get_required_forms(self):
        """
        Checks which forms are required based on the params and returns a list of only the required forms
        :rtype: list of NestedFormConfig
        """
        return [nested_form for nested_form in self.nested_forms if self.form_is_required(nested_form)]

    def form_is_required(self, nested_form):
        """
        Handles the logic to check if an individual form is required
        """

        # Check the required flag
        if nested_form.required:
            return True

        # Ensure we have a require key
        if nested_form.required_key is None:
            return False

        # Create a boolean field to ensure that the required key is truthy
        boolean_field = forms.BooleanField(required=False)

        # Return if this value is truthy or not
        return boolean_field.clean(self.data.get(nested_form.required_key))

    def _clean_fields(self):
        """
        Cleans all of self.data and populates self._errors and
        self.cleaned_data.

        This is copied from django with an addition at the bottom
        """

        # Clean our nested fields first
        # Get our required forms
        required_forms = self.get_required_forms()

        # Clean our required nested form fields first
        for form in required_forms:
            # Get the errors and cleaned data
            errors = form.instance.errors
            cleaned_data = form.instance.cleaned_data

            # Add the errors
            if errors:
                self.add_error(form.key, list(errors.values()))
            # Update the data if we do not have any errors
            else:
                self.data[form.key] = cleaned_data

        # Call the parent clean method to clean our base fields
        return super(NestedFormMixin, self)._clean_fields()



from django.apps import AppConfig


class RRuleTestsConfig(AppConfig):
    name = 'ambition_utils.rrule.tests'
    verbose_name = 'Ambition Utils RRule Tests'
    label = 'ambition_utils_rrule_tests'



from django import forms
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from unittest.mock import MagicMock

from ambition_utils.forms import NestedFormMixin, NestedFormConfig
from ambition_utils.tests.models import FakeModel


class BadFormNoSave(NestedFormMixin, forms.Form):
    pass


class NestedForm1(NestedFormMixin, forms.Form):
    three = forms.CharField(error_messages={
        'required': 'Three is required'
    })
    four = forms.CharField(required=False)

    def __init__(self, init_number, *args, **kwargs):
        self.init_number = init_number
        super(NestedForm1, self).__init__(*args, **kwargs)

    def save(self, number):
        return {
            'init_number': self.init_number,
            'number': number
        }


class NestedForm2(NestedFormMixin, forms.Form):
    five = forms.CharField(error_messages={
        'required': 'Five is required'
    })
    six = forms.CharField(required=False)

    def __init__(self, init_user_id, *args, **kwargs):
        self.init_user_id = init_user_id
        super(NestedForm2, self).__init__(*args, **kwargs)

    def save(self, user_id):
        return {
            'init_user_id': self.init_user_id,
            'user_id': user_id
        }


class OptionalForm(NestedFormMixin, forms.Form):
    seven = forms.CharField(error_messages={
        'required': 'seven is required'
    })
    eight = forms.CharField(required=False)

    def save(self):
        return 'optional value'


class FileInputNestedForm(NestedFormMixin, forms.Form):
    primary = forms.FileField(
        error_messages={'required': 'File is required'},
        required=True
    )

    backup = forms.FileField(
        required=False
    )

    def save(self, *args, **kwargs):
        return 'ABC'


class BadParentFormExistingField(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            pre=True,
        )
    ]

    # Duplicate form field from nested
    nested_form_1 = forms.CharField(error_messages={
        'required': 'Three is required'
    })

    def save(self):  # pragma: no cover
        pass


class ParentForm(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=False,
            required_key='nested_form_1_required',
            pre=True,
            error_messages={
                'three': {
                    'required': 'Nested three is required'
                },
                'bad': {
                    'required': 'Not required'
                }
            }
        ),
        NestedFormConfig(
            cls=NestedForm2,
            key='nested_form_2',
            required=False,
            required_key='nested_form_2_required',
            pre=True,
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            field_prefix='optional_1',
            required_key='optional_required',
            pre=True,
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional_2',
            required=False,
            field_prefix='optional_2',
            required_key='optional_required_2',
            post=True,
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)

    def get_nested_form_init_args(self, nested_form_config, args, kwargs, base_form_args, base_form_kwargs):
        if nested_form_config.key == 'nested_form_1':
            args = [1] + args
        elif nested_form_config.key == 'nested_form_2':
            args = [2] + args

        # Return the updated args
        return args, kwargs

    def get_nested_form_save_args(
        self,
        nested_form_config,
        args,
        kwargs,
        base_form_args,
        base_form_kwargs,
        form_responses
    ):
        if nested_form_config.key == 'nested_form_1':
            kwargs.update({
                'number': 10,
            })
        elif nested_form_config.key == 'nested_form_2':
            kwargs.update({
                'user_id': 5,
            })

        # Return the updated args
        return args, kwargs

    def save(self, *args, **kwargs):
        return 'the object'


class FormWithAlwaysRequired(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=True,
            pre=True
        ),
        NestedFormConfig(
            cls=NestedForm2,
            key='nested_form_2',
            required=True,
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=True,
            pre=True
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)


class FormWithOptional(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=NestedForm1,
            key='nested_form_1',
            required=False,
            required_key='nested_form_1_required',
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            required_key='optional_required',
            pre=True
        )
    ]

    one = forms.CharField(error_messages={
        'required': 'One is required'
    })
    two = forms.CharField(required=False)

    def get_nested_form_init_args(self, nested_form_config, args, kwargs, base_form_args, base_form_kwargs):
        if nested_form_config.key == 'nested_form_1':
            args = [1] + args

        # Return the updated args
        return args, kwargs

    def save(self):  # pragma: no cover
        return 'saved'


class ModelFormWithNestedForms(NestedFormMixin, forms.ModelForm):
    class Meta:
        model = FakeModel
        fields = ['name']

    nested_form_configs = [
        NestedFormConfig(
            cls=OptionalForm,
            key='optional',
            required=False,
            field_prefix='optional_1',
            required_key='optional_required',
            pre=True
        ),
        NestedFormConfig(
            cls=OptionalForm,
            key='optional_2',
            required=False,
            field_prefix='optional_2',
            required_key='optional_required_2',
            post=True
        )
    ]

    def save(self, commit=True, **kwargs):
        return super().save(commit=True)


class ParentOfFileInputs(NestedFormMixin, forms.Form):
    nested_form_configs = [
        NestedFormConfig(
            cls=FileInputNestedForm,
            key='file_input_nested_form',
            required=True,
            post=True,
        ),
    ]

    label = forms.CharField(error_messages={
        'required': 'Label is required'
    })

    def save(self, *args, **kwargs):
        return 'ABC'


class NestedFormMixinUnitTest(TestCase):
    """
    Unit tests for the nested form mixin
    """

    def test_bad_form_with_no_save_method(self):
        """
        Asssert that the form has a save method
        """

        with self.assertRaises(Exception):
            BadFormNoSave()

    def test_get_nested_form_init_args(self):
        """
        Checks return value of init args method
        """
        self.assertEqual(
            NestedFormMixin.get_nested_form_init_args(
                MagicMock(),
                nested_form_config=MagicMock(),
                args=['1'],
                kwargs={'one': 'one'},
                base_form_args=[],
                base_form_kwargs={}
            ),
            (['1'], {'one': 'one'})
        )

    def test_get_nested_form_save_args(self):
        """
        Checks return value of save args method
        """
        self.assertEqual(
            NestedFormMixin.get_nested_form_save_args(
                MagicMock(),
                nested_form_config=MagicMock(),
                args=['1'],
                kwargs={'one': 'one'},
                base_form_args=[],
                base_form_kwargs={},
                form_responses={}
            ),
            (['1'], {'one': 'one'})
        )

    def test_save(self):
        """
        Makes sure parent form save is called
        """
        class ParentForm1(forms.Form):
            def save(self):
                return 'saved'

        class ChildForm(NestedFormMixin, ParentForm1):
            pass

        form = ChildForm()
        self.assertEqual(form.save(), 'saved')

    def test_form_is_required(self):
        """
        Should return True when required is set, otherwise checks the flag value
        """

        # Create a test form
        class ParentForm1(NestedFormMixin, forms.Form):
            the_key = forms.BooleanField(required=False)

            def save(self):  # pragma: no cover
                return 'saved'

        # Required in config
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required=True
        )
        form = ParentForm1(data={})
        self.assertTrue(form.is_valid())
        self.assertTrue(form.form_is_required(nested_form))

        # Required from flag
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required_key='the_key'
        )
        form = ParentForm1(data={
            'the_key': '1'
        })
        self.assertTrue(form.is_valid())
        self.assertTrue(form.form_is_required(nested_form))

        # Not required from flag
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required_key='the_key'
        )
        form = ParentForm1(data={})
        self.assertTrue(form.is_valid())
        self.assertFalse(form.form_is_required(nested_form))

        # Required key is None
        nested_form = NestedFormConfig(
            cls=ParentForm1,
            key='one',
            required=False,
            required_key=None
        )
        form = ParentForm1(data={})
        self.assertTrue(form.is_valid())
        self.assertFalse(form.form_is_required(nested_form))


class NestedFormMixinTest(TestCase):
    """
    Tests for the nested form mixin with normal non model forms
    """

    def test_duplicate_form_field_from_nested(self):
        """
        Test a form that has a duplicated form field from a nested form
        """

        with self.assertRaises(Exception):
            BadParentFormExistingField()

    def test_field_prefix_validation_error(self):
        """
        Makes sure a validation error is raised if two of the same form exist but both don't have a field prefix
        """
        class Form1(forms.Form):
            pass

        class ParentForm1(NestedFormMixin, forms.Form):
            nested_form_configs = [
                NestedFormConfig(cls=Form1, key='one'),
                NestedFormConfig(cls=Form1, key='two')
            ]

            def save(self):  # pragma: no cover
                return 'saved'

        # Assert we raise a validation error from the two forms without a prefix
        with self.assertRaises(ValidationError):
            ParentForm1()

    def test_nested_required_fields(self):
        """
        The ParentForm required field and 2 of the nested forms' required fields should be required.
        The second optional form has a field prefix and provides the required field,
        so it should not cause an additional validation error.
        """

        # Setup the data
        data = {
            'nested_form_1_required': '1',
            'nested_form_2_required': '1',
            'optional_required_2': '1',
            'optional_2_seven': 'good'
        }

        # Create the form
        form = ParentForm(data=data)

        # Assert that the from is invalid
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors), 3)

        # Assert that we properly override our error message
        self.assertEqual(
            form.errors['nested_form_1'],
            ['Nested three is required']
        )

        # Make sure we didn't write over the base form class's error message
        form = NestedForm1(10)
        self.assertEqual(form.fields['three'].error_messages['required'], 'Three is required')

    def test_optional_form_flag_false(self):
        """
        Verifies an optional form is not required when its flag is false
        """
        data = {
            'nested_form_1_required': '1',
            'optional_required': 'False'
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors), 2)

    def test_optional_form_flag_missing(self):
        """
        Verifies an optional form is not required when its flag is missing
        """
        data = {
            'nested_form_1_required': '1',
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors), 2)

    def test_optional_form_flag_true(self):
        """
        Verifies an optional form is required when its flag is true
        """
        data = {
            'nested_form_1_required': 'True',
            'optional_required': 'True'
        }

        form = FormWithOptional(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors), 3)

    def test_full_scenario(self):
        """
        Covers all presave, save, postsave scenarios with multiple forms
        """

        # Setup the testing data
        data = {
            'nested_form_1_required': '1',
            'nested_form_2_required': '1',
            'optional_required_2': '1',
            'one': '1',
            'three': '3',
            'five': '5',
            'seven': '7',
        }

        # Create the form
        form = ParentForm(data=data)

        # Assert that the form is valid
        self.assertTrue(form.is_valid())

        # Save the form
        return_value = form.save()

        # Assert that we have the proper return value
        self.assertEqual(return_value, 'the object')


class NestedModelFormMixinTest(TestCase):

    def test_save(self):
        """
        Verifies the model save still works as expected
        """

        # Create the data to submit
        data = {
            'name': 'just a fake name',
            'optional_required': 'True',
            'optional_required_2': 'True',
            'seven': '7',
        }

        # Create the form
        form = ModelFormWithNestedForms(data=data)

        # Assert the form is valid
        self.assertTrue(form.is_valid())

        # Save the form
        form.save(commit=True)

        # Get the model that was saved
        fake_model = FakeModel.objects.get()

        # Assert the model was saved correctly
        self.assertEqual(fake_model.name, 'just a fake name')


class NestedFileInputFormTest(TestCase):

    def test_parent_validity(self):
        """
        Verifies that a form with nested forms containing FileField is valid when valid and not when not
        """

        data = {
            'label': 'Secret Plans',
        }

        files = {
            'primary': SimpleUploadedFile('primary_plan.txt', 'ssshhhh'.encode()),
            'backup': SimpleUploadedFile('primary_plan.txt', 'ssshhhh'.encode()),
        }

        # Create the form without files
        form = ParentOfFileInputs(data=data, files={})

        # Assert the form is not valid
        self.assertFalse(form.is_valid())

        # Create the form with files
        form = ParentOfFileInputs(data=data, files=files)

        # Assert the form is valid
        self.assertTrue(form.is_valid())

        # Save the form
        return_value = form.save()

        # Assert that we have the proper return value
        self.assertEqual(return_value, 'ABC')
